"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from codecs import open

# Always prefer setuptools over distutils
from setuptools import setup

# Get the long description from the README file
with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="pyserilog",

    version="0.0.9",

    description="python version of serilog a structured logging library",
    long_description=long_description,
    author='Reza Sadeghi',
    author_email='rezasadeghikhas@gmail.com',
    license="Apache2",

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: System :: Logging',
        'Topic :: Software Development :: Libraries',
        'Topic :: Internet :: Log Analysis',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.11',
    ],

    package_dir={"": "src"}

)



from pyserilog import LoggerConfiguration
from pyserilog.configuration.logger_destructuring_configuration import LoggerDestructuringConfiguration
from pyserilog.configuration.logger_enrichment_configuration import LoggerEnrichmentConfiguration
from pyserilog.configuration.logger_filter_configuration import LoggerFilterConfiguration
from pyserilog.configuration.logger_sink_configuration import LoggerSinkConfiguration
from pyserilog.configuration.logger_audit_sink_configuration import LoggerAuditSinkConfiguration
from pyserilog.core.ilog_event_filter import ILogEventFilter
from pyserilog.core.idestructuring_policy import IDestructuringPolicy
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.events.level_alias import LevelAlias
from pyserilog.events.log_event_level import LogEventLevel


class WriteToSurrogateConfigurationMethods:
    @staticmethod
    def sink(logger_sink_configuration: LoggerSinkConfiguration, sink: ILogEventSink,
             restricted_to_minimum_level: LogEventLevel = LevelAlias.minimum,
             level_switch: LoggingLevelSwitch = None) -> LoggerConfiguration:
        return logger_sink_configuration.sink(sink, restricted_to_minimum_level, level_switch)


class AuditToSurrogateConfigurationMethods:
    @staticmethod
    def sink(audit_sink_configuration: LoggerAuditSinkConfiguration, sink: ILogEventSink,
             restricted_to_minimum_level: LogEventLevel = LevelAlias.minimum,
             level_switch: LoggingLevelSwitch = None) -> LoggerConfiguration:
        return audit_sink_configuration.sink(sink, restricted_to_minimum_level, level_switch)


class EnrichSurrogateConfigurationMethods:
    @staticmethod
    def with_enrich(logger_enrichment_configuration: LoggerEnrichmentConfiguration,
                    enricher: ILogEventEnricher) -> LoggerConfiguration:
        return logger_enrichment_configuration.with_enrichers(enricher)


class SurrogateConfigurationMethods:

    @staticmethod
    def with_destructure(logger_destructuring_configuration: LoggerDestructuringConfiguration,
                         policy: IDestructuringPolicy) -> LoggerConfiguration:
        return logger_destructuring_configuration.with_policies(policy)

    @staticmethod
    def with_filter(logger_filter_configuration: LoggerFilterConfiguration, event_filter: ILogEventFilter):
        return logger_filter_configuration.with_filter(event_filter)

    @staticmethod
    def as_scalar(logger_destructuring_configuration: LoggerDestructuringConfiguration, scalar_type: type):
        return logger_destructuring_configuration.as_scalar(scalar_type)

    @staticmethod
    def to_maximum_depth(logger_destructuring_configuration: LoggerDestructuringConfiguration,
                         maximum_destructuring_depth: int):
        return logger_destructuring_configuration.to_maximum_depth(maximum_destructuring_depth)

    @staticmethod
    def to_maximum_string_length(logger_destructuring_configuration: LoggerDestructuringConfiguration,
                                 maximum_string_length: int):
        return logger_destructuring_configuration.to_maximum_string_length(maximum_string_length)

    @staticmethod
    def to_maximum_collection_count(logger_destructuring_configuration: LoggerDestructuringConfiguration,
                                    maximum_collection_count: int):
        return logger_destructuring_configuration.to_maximum_collection_count(maximum_collection_count)



from pyserilog.core.string_writable import StringIOWriteable,StringWriteable

from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events import message_template
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.formatting.json.json_value_formatter import JsonValueFormatter
from pyserilog.parsing.property_token import PropertyToken
from pyserilog.parsing.text_token import TextToken
from pyserilog.rendering.padding import Padding


class MessageTemplateRenderer:
    json_value_formatter: JsonValueFormatter = JsonValueFormatter("$type")

    @staticmethod
    def render(message_template: message_template, properties: dict[str, LogEventPropertyValue], output: StringWriteable,
               text_format: str = None):
        is_literal = False
        is_json = False

        if text_format is not None:
            if 'j' in text_format:
                is_json = True
            if 'l' in text_format:
                is_literal = True

        for token in message_template.tokens:
            if isinstance(token, TextToken):
                MessageTemplateRenderer.render_text_token(token , output)
            elif isinstance(token, PropertyToken):
                MessageTemplateRenderer.render_property_token(token, properties, output, is_literal, is_json)

    @staticmethod
    def render_text_token(token: TextToken, output: StringWriteable):
        output.write(token.text)

    @staticmethod
    def render_property_token(pt: PropertyToken, properties: dict[str, LogEventPropertyValue],
                              output: StringWriteable, is_literal: bool, is_json: bool):
        if pt.property_name not in properties:
            output.write(pt.raw_text)
            return

        property_value = properties[pt.property_name]
        if pt.alignment is None:
            MessageTemplateRenderer.render_value(property_value, is_literal, is_json, output, pt.format)
            return
        value_output = StringIOWriteable()
        MessageTemplateRenderer.render_value(property_value, is_literal, is_json, value_output, pt.format)
        sb = value_output.getvalue()
        if len(sb) >= pt.alignment.width:
            output.write(sb)
            return
        Padding.apply(output, sb, pt.alignment)

    @staticmethod
    def render_value(property_value: LogEventPropertyValue, is_literal: bool, is_json: bool, output: StringWriteable,
                     text_format: str | None = None):
        if is_literal and isinstance(property_value, ScalarValue) and isinstance(property_value.value, str):
            output.write(property_value.value)
        elif is_json and text_format is None:
            MessageTemplateRenderer.json_value_formatter.format(property_value, output)
        else:
            property_value.render(output, text_format)



from abc import ABC, abstractmethod

from pyserilog.core.ilog_event_property_factory import ILogEventPropertyFactory
from pyserilog.events.log_event import LogEvent


class ILogEventEnricher(ABC):

    @abstractmethod
    def enrich(self, log_event: LogEvent, property_factory: ILogEventPropertyFactory):
        pass



from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.guard import Guard
from pyserilog.events.event_property import EventProperty
from pyserilog.events.property_utils import ensure_valid_name


class LogEventProperty:

    def __init__(self, name: str, value: LogEventPropertyValue):
        Guard.against_null(name)
        ensure_valid_name(name)
        self._name = name
        self._value = value

    @staticmethod
    def create_with_event_property(event_property: EventProperty):
        return LogEventProperty(event_property.name, event_property.value)

    @property
    def value(self) -> LogEventPropertyValue:
        return self._value

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return f"(name = {self.name} , value = {self.value})"

    def __eq__(self, other):
        if isinstance(other, LogEventProperty):
            return self.name == other.name and self.value == other.value
        return False



from typing import Iterable

from pyserilog.capturing.depth_limiter import DepthLimiter
from pyserilog.capturing.iproperty_value_converter import IPropertyValueConverter
from pyserilog.core.idestructuring_policy import IDestructuringPolicy
from pyserilog.core.iscalar_conversion_policy import IScalarConversionPolicy
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.dictionary_value import DictionaryValue
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.events.sequence_value import SequenceValue
from pyserilog.events.structure_value import StructureValue
from pyserilog.parsing.destructuring import Destructuring
from pyserilog.policies.ByteArrayScalarConversionPolicy import ByteArrayScalarConversionPolicy
from pyserilog.policies.ReflectionTypesScalarDestructuringPolicy import ReflectionTypesScalarDestructuringPolicy
from pyserilog.policies.SimpleScalarConversionPolicy import SimpleScalarConversionPolicy
import pyserilog.utils as utils

built_in_scalar_types = [
    bool, int, float
    # chr,
    # str, datetime.date, decimal.Decimal
]


class PropertyValueConverter(IPropertyValueConverter):
    def __init__(self, maximum_destructuring_depth: int, maximum_string_length: int, maximum_collection_count: int,
                 additional_scalar_types: list[type], additional_destructuring_policies: list[IDestructuringPolicy],
                 propagate_exceptions: bool):
        self._maximum_string_length = maximum_string_length
        self._maximum_collection_count = maximum_collection_count
        self._propagate_exceptions = propagate_exceptions
        scalar_types = list(built_in_scalar_types)
        scalar_types.extend(additional_scalar_types)
        self._scalarConversionPolicies: list[IScalarConversionPolicy] = [
            SimpleScalarConversionPolicy(scalar_types),
            ByteArrayScalarConversionPolicy()
        ]

        self._destructuringPolicies = additional_destructuring_policies
        self._destructuringPolicies.append(ReflectionTypesScalarDestructuringPolicy())
        self._depthLimiter = DepthLimiter(maximum_destructuring_depth, self)

    def create_property(self, name: str, value, destructure_objects: bool = False) -> LogEventProperty:
        prop = self.create_property_value_check_destructuring(value, destructure_objects)
        return LogEventProperty(name, prop)

    def create_property_value_check_destructuring(self, value,
                                                  destructure_objects: bool = False) -> LogEventPropertyValue:
        return self.create_property_value_check_destructuring_and_depth(value, destructure_objects, 1)

    def create_property_value_with_destructuring(self, value: object,
                                                 destructuring: Destructuring) -> LogEventPropertyValue:
        try:
            return self.create_property_value_with_destructuring_and_depth(value, destructuring, 1)
        except Exception as ex:
            if self._propagate_exceptions:
                raise ex
            SelfLog.write_line("Exception caught while converting property value: {0}", ex)
            return ScalarValue(f"Capturing the property value threw an exception: {utils.full_name(ex)}")

    def create_property_value_check_destructuring_and_depth(self,
                                                            value, destructure_objects: bool,
                                                            depth: int) -> LogEventPropertyValue:
        destructuring = Destructuring.DESTRUCTURE if destructure_objects == True else Destructuring.DEFAULT
        return self.create_property_value_with_destructuring_and_depth(value, destructuring, depth)

    def create_property_value_with_destructuring_and_depth(self, value, destructuring: Destructuring,
                                                           depth: int) -> LogEventPropertyValue:
        if value is None:
            return ScalarValue.Null()

        if destructuring == Destructuring.STRINGIFY:
            return self.__stringify(value)

        if destructuring == Destructuring.DESTRUCTURE and isinstance(value, str):
            value = self.__truncate_if_necessary(value)

        if isinstance(value, str):
            return ScalarValue(value)

        for converter in self._scalarConversionPolicies:
            checked, res = converter.try_convert_to_scalar(value)
            if checked:
                return res

        DepthLimiter.set_current_depth(depth)

        if destructuring == Destructuring.DESTRUCTURE:
            for destructuring_policy in self._destructuringPolicies:
                checked, res = destructuring_policy.try_destructure(value, self._depthLimiter)
                if checked:
                    return res
        if isinstance(value, Iterable) or hasattr(value, '__iter__'):
            return self.__convert_iterables(value, destructuring)

        checked, res = self.__try_convert_compiler_generated_type(value, destructuring)
        if checked:
            return res
        return ScalarValue(str(value))

    def __stringify(self, value) -> ScalarValue:
        if value is None:
            return ScalarValue("")
        value_str: str = str(value)
        truncated = self.__truncate_if_necessary(value_str)
        return ScalarValue(truncated)

    def __truncate_if_necessary(self, text: str):
        if len(text) > self._maximum_string_length:
            return f"{text[0:self._maximum_string_length - 1]}…"
        return text

    def __convert_iterables(self, value, destructuring: Destructuring):
        # Only dictionaries with 'scalar' keys are permitted, as
        # more complex keys may not serialize to unique values for
        # representation in sinks. This check strengthens the expectation
        # that resulting dictionary is representable in JSON as well
        # as richer formats (e.g. XML, .NET type-aware...).
        # Only actual dictionaries are supported, as arbitrary types
        # can implement multiple IDictionary interfaces and thus introduce
        # multiple different interpretations.

        if isinstance(value, dict):
            def map_to_dictionary_elements():
                result: list = []
                count = 0
                for key in value:
                    if count >= self._maximum_collection_count:
                        break
                    count += 1
                    val = value[key]
                    key_value = self._depthLimiter.create_property_value_with_destructure(key, destructuring)
                    val_value = self._depthLimiter.create_property_value_with_destructure(val, destructuring)
                    result.append((key_value, val_value))
                return result

            elements = map_to_dictionary_elements()
            res_dict = DictionaryValue(elements)
            return res_dict

        if len(value) <= self._maximum_collection_count:
            if len(value) == 0:
                return SequenceValue.empty()
            arr = []
            for el in value:
                element_value = self._depthLimiter.create_property_value_with_destructure(el, destructuring)
                arr.append(element_value)
            return SequenceValue(arr)
        else:
            seq = self.__map_to_sequence(value, destructuring)
            return SequenceValue(seq)

    def __map_to_sequence(self, sequence: list, destructure: Destructuring):
        count = 0
        res: list[LogEventPropertyValue] = []
        for element in sequence:
            if count < self._maximum_collection_count:
                val = self._depthLimiter.create_property_value_with_destructure(element, destructure)
                res.append(val)
                count += 1
            else:
                break
        return res

    def __try_convert_compiler_generated_type(self, value, destructuring: Destructuring) \
            -> tuple[bool, LogEventPropertyValue | None]:
        if destructuring == Destructuring.DESTRUCTURE:
            type_tag = type(value).__name__
            if len(type_tag) <= 0:
                type_tag = None
            props: list[LogEventProperty] = self.__get_properties(value)
            return True, StructureValue(props, type_tag)
        return False, None

    def __get_properties(self, value) -> list[LogEventProperty]:
        res: list[LogEventProperty] = []
        public_props = list(name for name in dir(value) if not name.startswith('_'))
        for property_name in public_props:
            try:
                property_value = getattr(value, property_name)
            except Exception as ex:
                SelfLog.write_line("The property accessor {0} threw exception: {1}", property_name, ex)
                if self._propagate_exceptions:
                    raise ex
                property_value = f"The property accessor threw an exception: {type(ex).__name__}"
            prop_value = self._depthLimiter.create_property_value_with_destructure(
                property_value, Destructuring.DESTRUCTURE)
            event_prop = LogEventProperty(property_name, prop_value)
            res.append(event_prop)
        return res



from typing import Callable

from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.ilog_event_property_factory import ILogEventPropertyFactory
from pyserilog.events.log_event import LogEvent


class ConditionalEnricher(ILogEventEnricher):
    def __init__(self, wrapped: ILogEventEnricher, condition: Callable[[LogEvent], bool]):
        self._wrapped = wrapped
        self._condition = condition

    def enrich(self, log_event: LogEvent, property_factory: ILogEventPropertyFactory):
        if self._condition(log_event):
            self._wrapped.enrich(log_event, property_factory)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._wrapped, "__exit__"):
            self._wrapped.__exit__(exc_type, exc_val, exc_tb)



from typing import Callable

from pyserilog.configuration.logger_sink_configuration import LoggerSinkConfiguration
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.level_alias import LevelAlias


class LoggerAuditSinkConfiguration:

    def __init__(self, logger_configuration, add_sink_function):
        self._sink_configuration: LoggerSinkConfiguration = \
            LoggerSinkConfiguration(logger_configuration, add_sink_function)

    def sink(self, log_event_sink: ILogEventSink, restricted_to_minimum_level: LogEventLevel = LevelAlias.minimum,
             level_switch: LoggingLevelSwitch | None = None):
        return self._sink_configuration.sink(log_event_sink, restricted_to_minimum_level, level_switch)

    def logger(self, configure_logger_function: Callable,
               restricted_to_minimum_level: LogEventLevel = LevelAlias.minimum,
               level_switch: LoggingLevelSwitch | None = None
               ):
        return self._sink_configuration.logger(configure_logger_function, restricted_to_minimum_level, level_switch)



from abc import ABC, abstractmethod

from pyserilog.events.scalar_value import ScalarValue


class IScalarConversionPolicy(ABC):
    @abstractmethod
    def try_convert_to_scalar(self, value) -> tuple[bool, ScalarValue | None]:
        pass



from typing import Callable

from pyserilog.guard import Guard
from pyserilog.core.ilog_event_filter import ILogEventFilter
from pyserilog.core.filter.delegate_filter import DelegateFilter
from pyserilog.events.log_event import LogEvent


class LoggerFilterConfiguration:

    def __init__(self, logger_configuration, add_filter: Callable[[ILogEventFilter], None]):
        self._logger_configuration = Guard.against_null(logger_configuration)
        self._add_filter: Callable[[ILogEventFilter], None] = Guard.against_null(add_filter)

    def with_filter(self, *filters):
        """
        Filter out log events from the stream based on the provided filter.
        :param filters: The filters to apply.
        :return: Configuration object allowing method chaining.
        """
        for log_event_filer in filters:
            if log_event_filer is None:
                raise ValueError("Null filter is not allowed.")
            self._add_filter(log_event_filer)
        return self._logger_configuration

    def by_excluding(self, exclusion_predicate: Callable[[LogEvent], bool]):

        def is_enable_func(log_level: LogEvent):
            res = not exclusion_predicate(log_level)
            return res

        return self.with_filter(DelegateFilter(is_enable_func=is_enable_func))

    def by_including_only(self, inclusion_predicate: Callable[[LogEvent], bool]):
        return self.with_filter(DelegateFilter(inclusion_predicate))



class Casing:

    @staticmethod
    def format(value: str, text_format: str | None):
        match text_format:
            case "u":
                return value.upper()
            case "w":
                return value.lower()
            case _:
                return value



def full_name(class_type: type):
    if not isinstance(class_type, type):
        class_type = type(class_type)
    module = class_type.__module__
    if module == "__builtin__" or module == 'builtins':
        return class_type.__qualname__
    return f'{module}.{class_type.__qualname__}'



import threading

from pyserilog.guard import Guard
from pyserilog.core.imessage_template_parser import IMessageTemplateParser
from pyserilog.events.message_template import MessageTemplate


class MessageTemplateCache(IMessageTemplateParser):
    max_cache_items = 1000,
    max_cached_template_length = 1024
    _template_lock = threading.Lock()
    _templates = {}

    def __init__(self, inner_parser: IMessageTemplateParser):
        self._inner_parser = inner_parser

    def parse(self, message_template: str) -> MessageTemplate:
        Guard.against_null(message_template)

        if len(message_template) > self.max_cached_template_length:
            return self._inner_parser.parse(message_template)

        if message_template in self._templates:
            return self._templates[message_template]

        result = self._inner_parser.parse(message_template)
        with self._template_lock:
            """
            Exceeding MaxCacheItems is *not* the sunny day scenario; all we're doing here is preventing out-of-memory
            conditions when the library is used incorrectly. Correct use (templates, rather than
            direct message strings) should barely, if ever, overflow this cache.
            
            Changing workloads through the lifecycle of an app instance mean we can gain some ground by
            potentially dropping templates generated only in startup, or only during specific infrequent activities.
            """

            if len(self._templates) == self.max_cache_items:
                self._templates.clear()
            self._templates[message_template] = result

        return result



from abc import ABC, abstractmethod
from pyserilog.core.string_writable import StringIOWriteable,StringWriteable
from typing import Callable


class LogEventPropertyValue(ABC):

    @abstractmethod
    def render(self, output: StringWriteable, text_format: str | None = None, formatter: Callable[[object], str] = None):
        pass

    def __str__(self) -> str:
        string_writer = StringIOWriteable()
        self.render(string_writer)
        res = string_writer.getvalue()
        string_writer.close()
        return res
